--[[
	Core module, 
	this module is loaded automatically
]]
fw_addModule('fw',[[
-- general debug flag
debug = {
	enabled = false
}
hooks = {}
hooks.doors = {}
hooks.monsters = {}
hooks.party = {}
hooks.items = {}

hookOrder = {}
hookVars = {}
currentHook = ''
repeatingTimers = {}

lists = {}

function debugPrint(message)
	if fw.debug.enabled then print(message) end
end

function debugPrintHookOrder(namespace)
	for i,v in ipairs(hookOrder[namespace]) do
		fw.debugPrint(i..":"..v)
	end
end

function addHooks(hookNamespace,hooksId,hookFunctions,ordinal)
	if (not fw.hooks[hookNamespace]) then
		fw.hooks[hookNamespace] = {}
	end
	if (fw.hooks[hookNamespace][hooksId]) then 
		fw.debugPrint("Hooks "..hookNamespace..'.'..hooksId..' already defined')
		return false
	end
	if not fw.hookOrder[hookNamespace] then
		 fw.hookOrder[hookNamespace] = {}
	end
	if (ordinal) then
		table.insert(fw.hookOrder[hookNamespace],ordinal,hooksId)
	else
		table.insert(fw.hookOrder[hookNamespace],hooksId)
	end
	fw.hooks[hookNamespace][hooksId] = hookFunctions

end

function setHook(hookIdString,func,ordinal)
	local parts = help.split(hookIdString,'.')
	local hookNamespace = parts[1]
	local hooksId = parts[2]
	local hookFuncName = parts[3]
	if (not fw.hooks[hookNamespace]) then
		fw.hooks[hookNamespace] = {}
	end
	if (not fw.hooks[hookNamespace][hooksId]) then
		fw.hooks[hookNamespace][hooksId] = {}
	end	
	if fw.hooks[hookNamespace][hooksId][hookFuncName] then
		fw.debugPrint('Warning: Hook '..hookIdString..' was owerwritten')
	else
		setHookOrder(hookNamespace,hooksId,ordinal)
	end
	fw.hooks[hookNamespace][hooksId][hookFuncName] = func
end

function hooksExists(hookNamespace,hooksId)
	if (fw.hooks[hookNamespace] and fw.hooks[hookNamespace][hooksId]) then
		return true
	end
	return false
end


function removeHook(hookIdString)
	hookIdString = hookIdString or currentHook
	local parts = help.split(hookIdString,'.')
	local hookNamespace = parts[1]
	local hooksId = parts[2]
	local hookFuncName = parts[3]
	if fw.hooks[hookNamespace] and fw.hooks[hookNamespace][hooksId] then
		fw.hooks[hookNamespace][hooksId][hookFuncName] = nil
		if #fw.hooks[hookNamespace][hooksId] == 0 then
			removeHooks(hookNamespace,hooksId)
		end
	end
end

function setHookOrder(hookNamespace,hooksId,ordinal)
	if not fw.hookOrder[hookNamespace] then
		 fw.hookOrder[hookNamespace] = {}
	end	
	if (ordinal) then
		table.insert(fw.hookOrder[hookNamespace],ordinal,hooksId)
	else
		table.insert(fw.hookOrder[hookNamespace],hooksId)
	end	
end




function getId(entityOrChampion)
	if entityOrChampion.getOrdinal then
		return 'champion_'..entityOrChampion:getOrdinal()
	end
	return entityOrChampion.id
end

function getById(id)
	local entity = findEntity(id)
	if not entity and string.sub(id,1,9) == 'champion_' then
		for i=1,4 do
			local champ = party:getChampion(i)
			if champ and getId(champ) == id then
				return champ 
			end
		end
	end
	return entity
end


function setHookVars(hookNamespace,hooksId,hookFunction,vars)
	hookVars[hookNamespace..'.'..hooksId..'.'..hookFunction] = vars
end

function getHookVars()
	return hookVars[currentHook]
end

function removeHooks(hookNamespace,hooksId)
	if (not fw.hooks[hookNamespace]) then return end
	removeHookOrder(hookNamespace,hooksId)
	fw.hooks[hookNamespace][hooksId] = nil
end

function removeHookOrder(hookNamespace,hooksId)
	for i,v in pairs(fw.hookOrder[hookNamespace]) do
	  if v == hooksId then
		table.remove(fw.hookOrder[hookNamespace],i)
		break
	  end
	end
end



function executeHooks(hookNamespace,methodName,entity,p1,p2,p3,p4)
	if not fw.hooks[hookNamespace] then return end
	local returnValue = nil
	local finalReturnValue = nil
	if fw.hookOrder[hookNamespace] then
		local id = ''
		for i=1, # fw.hookOrder[hookNamespace] do
			id = fw.hookOrder[hookNamespace][i]
			if fw.hooks[hookNamespace][id] then
				
				currentHook = hookNamespace..'.'..id..'.'..methodName
				
				local hookFunction = fw.hooks[hookNamespace][id][methodName]
				if type(hookFunction) == "function" then 
					--fw.debugPrint("Hook "..hookNamespace..'.'..id..'.'..methodName..' executed')
					returnValue = hookFunction(entity,p1,p2,p3,p4)
					if returnValue == false then return false end
					if returnValue == true then finalReturnValue = true end
				end
			end
		end
	end
	return finalReturnValue
end

function executeEntityHooks(entityType,methodName,entity,p1,p2,p3)
	local finalRetVal = executeHooks(entity.id,methodName,entity,p1,p2,p3)
	if finalRetVal == false then return false end
	
	local retVal = executeHooks(entity.name,methodName,entity,p1,p2,p3)
	if retVal == false then return false end
	if retVal ~= nil then finalRetVal = retVal end
	
	retVal = executeHooks(entityType,methodName,entity,p1,p2,p3)
	if retVal ~= nil then finalRetVal = retVal end
	
	return finalRetVal
end		

-- DEPRECATED
function repeatFunction(count,interval,callbackargs,callback,instant)
	print('fw.repeatFunction is deprecated, use timers.repeatCall(interval,count,instant,callback,args) instead')
	timers.repeatCall(interval,count,instant,callback,callbackargs)
end

function delayCall(interval,callback,callbackargs)
	print('fw.delayCall is deprecated, use timers.delaycall(interval,callback,callbackargs) instead')
	timers.delaycall(interval,callback,callbackargs)
end

]])

local lists = {}

lists.monster = {
--[[	"crab","crowern","cube","goromorg","green_slime","herder","herder_big",
	"herder_small","herder_swarm","ice_lizard","ogre","scavenger","scavenger_swarm",
	"shrakk_torr","skeleton_archer","skeleton_archer_patrol","skeleton_patrol",
	"skeleton_warrior","snail","spider","tentacles","uggardian","warden","wyvern"]]
}



lists.door = {}
lists.item = {
	weapon = {},
	shield = {},
	accessory = {},
	bomb = {},
	food = {},
	missileweapon = {},
	tome = {},
	armor = {},
	cloth = {},
	herb = {},
	machinepart = {},
	potion = {},
	staff = {},
	treasure = {},
	container = {},
	key = {},
	miscitem = {},
	scroll = {},
	throwingweapon = {},
	consumable = {},
	custom = {}
}	
lists.button = {}
lists.alcove = {}
lists.blockage = {}
lists.spell = {}


function fw_getHooks(entityClass,itemSubtype)
	entityClass = string.lower(entityClass)
	if entityClass == 'item' then
		itemSubtype = itemSubtype or 'custom'
		return {
			onUseItem = function(self,p1,p2,p3)
				local retVal = fw.executeHooks('items_'..itemSubtype,'onUseItem',self,p1,p2,p3)
				if retVal == false then return false end
				retVal = fw.executeEntityHooks("items","onUseItem",self,p1,p2,p3)
				if retVal == false then return false end
				if retVal == nil and fw.lists.item.consumable[self.name] then
					return true
				end
				return retVal
			end,
			onEquipItem = function(champion, slot)
				-- check that the fw script is loaded (in editor this is called before the fw is loaded because the everburning torch is spawned to the inventory instantly)
				if not fw or fw.executeEntityHooks == nil then return end 
				local item = champion:getItem(slot)
				local retVal = fw.executeEntityHooks("items","onEquipItem",item,champion,slot)
				return retVal
			end,
			onUnequipItem = function(champion, slot)
				if not fw or fw.executeEntityHooks == nil then return end 
				local item = champion:getItem(slot)
				local retVal = fw.executeEntityHooks("items","onUnequipItem",item,champion,slot)
				return retVal
			end,	
		}
	end
	if entityClass == 'monster' then
		return {
			onMove = function(self,p1,p2,p3) return fw.executeEntityHooks("monsters","onMove",self,p1,p2,p3) end,
			onTurn = function(self,p1,p2,p3) return fw.executeEntityHooks("monsters","onTurn",self,p1,p2,p3) end,
			onAttack = function(self,p1,p2,p3) return fw.executeEntityHooks("monsters","onAttack",self,p1,p2,p3) end,
			onRangedAttack = function(self,p1,p2,p3) return fw.executeEntityHooks("monsters","onRangedAttack",self,p1,p2,p3) end,
			onDamage = function(self,p1,p2,p3) return fw.executeEntityHooks("monsters","onDamage",self,p1,p2,p3) end,
			onDealDamage = function(self,p1,p2,p3) return fw.executeEntityHooks("monsters","onDealDamage",self,p1,p2,p3) end,	
			onDie = function(self,p1,p2,p3) return fw.executeEntityHooks("monsters","onDie",self,p1,p2,p3) end,
			onProjectileHit = function(monster,projectile,damage,damageType) 
				if fw.executeEntityHooks("projectiles","onHit",projectile,monster,damage,damageType) == false then return false end
				return fw.executeEntityHooks("monsters","onProjectileHit",monster,projectile,damage,damageType) 
			end
		}
	end	
	if entityClass == 'door' then
		return {
			onClose = function(self,p1,p2,p3) return fw.executeEntityHooks("doors","onClose",self,p1,p2,p3) end,
			onOpen = function(self,p1,p2,p3) return fw.executeEntityHooks("doors","onOpen",self,p1,p2,p3) end,		
		}	
	end
	if entityClass == 'blockage' then	
		return {
			onDamage = function(self,p1,p2,p3) return fw.executeEntityHooks("blockages","onDamage",self,p1,p2,p3) end,	
			onProjectileHit = function(blockage,projectile,damage,damageType) 
				if fw.executeEntityHooks("projectiles","onHit",blockage,champion,damage,damageType) == false then return false end
				return fw.executeEntityHooks("blockages","onProjectileHit",blockage,projectile,damage,damageType) 
			end,
			onDie = function(self,p1,p2,p3) return fw.executeEntityHooks("blockages","onDie",self,p1,p2,p3) end,	
		}	
	end		
	if entityClass == 'alcove' then	
		return {
			onInsertItem = function(self,p1,p2,p3) 
				if fw.executeEntityHooks("alcoves","onInsertItem",self,p1,p2,p3) == false then return false  end
				return true
			end,		
		}	
	end			
	return {}
end


function fw_defineObject(def)
	local itemSubtype = def.itemCategory or 'custom'
	def.itemCategory = nil
	local hooks = fw_getHooks(def.class,itemSubtype)
	for hookName,hookFunc in pairs(hooks) do
	
		if def[hookName] == nil then
			def[hookName] = hookFunc
		end
	end
	-- Container:onUseItem not supported
	if def.container then
		def.onUseItem = nil	
	end
	
	if entityType == 'item' then	
		-- add item to the items list by subtype
		lists.item[itemSubtype][#lists.item[itemSubtype]] = entityName
		
	end		
	defineObject(def)
end

function fw_addHooks(entityName,entityType,itemSubtype)
	local def = fw_getHooks(entityType,itemSubtype)
	def.name = entityName
	def.baseObject = entityName
	cloneObject(def)	
end


for _,monsterName in pairs(lists.monster) do 
	fw_addHooks(monsterName,'monster')
end

for _,doorName in pairs(lists.door) do 
	fw_addHooks(doorName,'door')
end

for _,blockageName in pairs(lists.blockage) do 
	fw_addHooks(blockageName,'blockage')
end

for _,alcoveName in pairs(lists.alcove) do 
	fw_addHooks(alcoveName,'alcove')
end


for category,list in pairs(lists.item) do
    -- containers and food disabled for now because their behaviour seems to be hard coded 
	if (category ~= 'container' and category ~= 'food' and category ~= 'consumable') then
		for i=1,# list do 
			fw_addHooks(list[i],'item',category)
		end
	end
end

for i=1,# lists.alcove do 
	fw_addHooks(lists.alcove[i],'alcove')
end

for i=1,# lists.blockage do 
	fw_addHooks(lists.blockage[i],'blockage')
end

-- Party hooks
cloneObject{
	name = "party",
	baseObject = "party",
	onAttack = function(self,p1,p2,p3) 
		if fw.executeHooks("champion_"..self:getOrdinal(),"onAttack",self,p1,p2,p3) == false then return false end	
		return fw.executeHooks("party","onAttack",self,p1,p2,p3) 
	end,
	onDamage = function(self,p1,p2,p3)
		if fw.executeHooks("champion_"..self:getOrdinal(),"onDamage",self,p1,p2,p3) == false then return false end	
		return fw.executeHooks("party","onDamage",self,p1,p2,p3) 
	end,
	onDie = function(self,p1,p2,p3) 
		if fw.executeHooks("champion_"..self:getOrdinal(),"onDie",self,p1,p2,p3) == false then return false end
		return fw.executeHooks("party","onDie",self,p1,p2,p3) 
	end,
	onProjectileHit = function(champion,projectile,damage,damageType) 
		if fw.executeEntityHooks("projectiles","onHit",projectile,champion,damage,damageType)  == false then return false end
		if fw.executeHooks("champion_"..champion:getOrdinal(),"onProjectileHit",champion,projectile,damage,damageType) == false then return false end
		return fw.executeHooks("party","onProjectileHit",champion,projectile,damage,damageType) 
	end,
	onUseItem = function(self,p1,p2,p3) 
		if fw.executeHooks("champion_"..self:getOrdinal(),"onUseItem",self,p1,p2,p3) == false then return false end
		return fw.executeHooks("party","onUseItem",self,p1,p2,p3) 
	end,
	onReceiveCondition = function(self,p1,p2,p3) 
		if fw.executeHooks("champion_"..self:getOrdinal(),"onReceiveCondition",self,p1,p2,p3) == false then return false end
		return fw.executeHooks("party","onReceiveCondition",self,p1,p2,p3) 
	end,
	onCastSpell = function(self,p1,p2,p3) 
		if fw.executeHooks("champion_"..self:getOrdinal(),"onCastSpell",self,p1,p2,p3) == false then return false end
		return fw.executeHooks("party","onCastSpell",self,p1,p2,p3) 
	end,
	onLevelUp = function(self,p1,p2,p3) 
		if fw.executeHooks("champion_"..self:getOrdinal(),"onLevelUp",self,p1,p2,p3) == false then return false end
		return fw.executeHooks("party","onLevelUp",self,p1,p2,p3) 
	end,
	onPickUpItem = function(self,p1,p2,p3) return fw.executeHooks("party","onPickUpItem",self,p1,p2,p3) end,
	onRest = function(self,p1,p2,p3) return fw.executeHooks("party","onRest",self,p1,p2,p3) end,	
	onWakeUp = function(self,p1,p2,p3) return fw.executeHooks("party","onWakeUp",self,p1,p2,p3) end,	
	onMove = function(self,p1,p2,p3) return fw.executeHooks("party","onMove",self,p1,p2,p3) end,
	onTurn = function(self,p1,p2,p3) return fw.executeHooks("party","onTurn",self,p1,p2,p3) end,	
	onDrawGui = function(g) fw.executeHooks("party","onDrawGui",g) end,
	onDrawInventory = function(g,champion) return fw.executeHooks("party","onDrawInventory",g,champion) end,
	onDrawStats = function(g,champion) return fw.executeHooks("party","onDrawStats",g,champion) end,
	onDrawSkills = function(g,champion) return fw.executeHooks("party","onDrawSkills",g,champion) end,
}

-- this function is called automatically after the fw script entity is initialized
fw_addModuleInitCallback('fw',function(fw,grimq)
		fw.lists['monster'] = tableToSet(lists.monster)
		fw.lists['door'] = tableToSet(lists.door)
		fw.lists['button'] = tableToSet(lists.button)
		fw.lists['blockage'] = tableToSet(lists.blockage)
		fw.lists['item'] = {}
		for i,v in pairs(lists.item) do
			fw.lists['item'][i] = tableToSet(v)
		end
		grimq.fromAllEntitiesInWorld()
			:foreach(function(e)
					if grimq.isMonster(e) then
						fw.lists['monster'][e.name] = true
					end
					if grimq.isItem(e) then
						local found = false
						for itemType, itemList in pairs(fw.lists.item) do
							if itemList[e.name] then 
								found = true
								break
							end
						end
						if not found then
							fw.lists['item']['custom'][e.name] = true
						end
					end		
					if grimq.isDoor(e) then
						fw.lists['door'][e.name] = true
					end			
				end
			)		
end
)




